@@ -1,20 +1,37 @@
 use super::{Arch, Os};
 use crate::{
     error::{OciSpecError, Result},
-    from_file, from_reader, to_file, to_writer,
+    from_file, from_reader, to_file, to_string, to_writer,
 };
 use derive_builder::Builder;
-use getset::{Getters, Setters};
+use getset::{CopyGetters, Getters, MutGetters, Setters};
 use serde::{ser::SerializeMap, Deserialize, Deserializer, Serialize, Serializer};
 #[cfg(test)]
 use std::collections::BTreeMap;
 use std::{
     collections::HashMap,
+    fmt::Display,
     io::{Read, Write},
     path::Path,
 };
 
-#[derive(Builder, Clone, Debug, Deserialize, Eq, Getters, Setters, PartialEq, Serialize)]
+/// In theory, this key is not standard.  In practice, it's used by at least the
+/// RHEL UBI images for a long time.
+pub const LABEL_VERSION: &str = "version";
+
+#[derive(
+    Builder,
+    Clone,
+    Debug,
+    Default,
+    Deserialize,
+    Eq,
+    Getters,
+    MutGetters,
+    Setters,
+    PartialEq,
+    Serialize,
+)]
 #[builder(
     default,
     pattern = "owned",
@@ -73,10 +90,13 @@ pub struct ImageConfiguration {
     /// The rootfs key references the layer content addresses used by the
     /// image. This makes the image config hash depend on the
     /// filesystem hash.
+    #[getset(get_mut = "pub", get = "pub", set = "pub")]
     rootfs: RootFs,
     /// Describes the history of each layer. The array is ordered from first
     /// to last.
-    history: Vec<History>,
+    #[serde(skip_serializing_if = "Option::is_none")]
+    #[getset(get_mut = "pub", get = "pub", set = "pub")]
+    history: Option<Vec<History>>,
 }
 
 impl ImageConfiguration {
@@ -175,27 +195,91 @@ impl ImageConfiguration {
     pub fn to_writer_pretty<W: Write>(&self, writer: &mut W) -> Result<()> {
         to_writer(&self, writer, true)
     }
+
+    /// Attempts to write an image configuration to a string as JSON.
+    /// # Errors
+    /// This function will return an [OciSpecError::SerDe](crate::OciSpecError::SerDe) if
+    /// the image configuration cannot be serialized.
+    /// # Example
+    /// ``` no_run
+    /// use oci_spec::image::ImageConfiguration;
+    ///
+    /// let image_configuration = ImageConfiguration::from_file("config.json").unwrap();
+    /// let json_str = image_configuration.to_string().unwrap();
+    /// ```
+    pub fn to_string(&self) -> Result<String> {
+        to_string(&self, false)
+    }
+
+    /// Attempts to write an image configuration to a string as pretty printed JSON.
+    /// # Errors
+    /// This function will return an [OciSpecError::SerDe](crate::OciSpecError::SerDe) if
+    /// the image configuration cannot be serialized.
+    /// # Example
+    /// ``` no_run
+    /// use oci_spec::image::ImageConfiguration;
+    ///
+    /// let image_configuration = ImageConfiguration::from_file("config.json").unwrap();
+    /// let json_str = image_configuration.to_string_pretty().unwrap();
+    /// ```
+    pub fn to_string_pretty(&self) -> Result<String> {
+        to_string(&self, true)
+    }
+
+    /// Extract the labels of the configuration, if present.
+    pub fn labels_of_config(&self) -> Option<&HashMap<String, String>> {
+        self.config().as_ref().and_then(|c| c.labels().as_ref())
+    }
+
+    /// Retrieve the version number associated with this configuration.  This will try
+    /// to use several well-known label keys.
+    pub fn version(&self) -> Option<&str> {
+        let labels = self.labels_of_config();
+        if let Some(labels) = labels {
+            for k in [super::ANNOTATION_VERSION, LABEL_VERSION] {
+                if let Some(v) = labels.get(k) {
+                    return Some(v.as_str());
+                }
+            }
+        }
+        None
+    }
+
+    /// Extract the value of a given annotation on the configuration, if present.
+    pub fn get_config_annotation(&self, key: &str) -> Option<&str> {
+        self.labels_of_config()
+            .and_then(|v| v.get(key).map(|s| s.as_str()))
+    }
 }
 
-impl Default for ImageConfiguration {
-    fn default() -> Self {
-        Self {
-            created: Default::default(),
-            author: Default::default(),
-            architecture: Arch::Amd64,
-            os: Os::Linux,
-            os_version: Default::default(),
-            os_features: Default::default(),
-            variant: Default::default(),
-            config: Default::default(),
-            rootfs: Default::default(),
-            history: Default::default(),
-        }
+/// This ToString trait is automatically implemented for any type which implements the Display trait.
+/// As such, ToString shouldn’t be implemented directly: Display should be implemented instead,
+/// and you get the ToString implementation for free.
+impl Display for ImageConfiguration {
+    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
+        // Serde serialization never fails since this is
+        // a combination of String and enums.
+        write!(
+            f,
+            "{}",
+            self.to_string_pretty()
+                .expect("ImageConfiguration JSON conversion failed")
+        )
     }
 }
 
 #[derive(
-    Builder, Clone, Debug, Default, Deserialize, Eq, Getters, Setters, PartialEq, Serialize,
+    Builder,
+    Clone,
+    Debug,
+    Default,
+    Deserialize,
+    Eq,
+    Getters,
+    MutGetters,
+    Setters,
+    PartialEq,
+    Serialize,
 )]
 #[serde(rename_all = "PascalCase")]
 #[builder(
@@ -268,6 +352,7 @@ pub struct Config {
     /// The field contains arbitrary metadata for the container.
     /// This property MUST use the annotation rules.
     #[serde(skip_serializing_if = "Option::is_none")]
+    #[getset(get_mut = "pub", get = "pub", set = "pub")]
     labels: Option<HashMap<String, String>>,
     /// The field contains the system call signal that will be
     /// sent to the container to exit. The signal can be a signal
@@ -331,7 +416,9 @@ where
     }
 }
 
-#[derive(Builder, Clone, Debug, Deserialize, Eq, Getters, Setters, PartialEq, Serialize)]
+#[derive(
+    Builder, Clone, Debug, Deserialize, Eq, Getters, MutGetters, Setters, PartialEq, Serialize,
+)]
 #[builder(
     default,
     pattern = "owned",
@@ -346,6 +433,7 @@ pub struct RootFs {
     typ: String,
     /// An array of layer content hashes (DiffIDs), in order
     /// from first to last.
+    #[getset(get_mut = "pub", get = "pub", set = "pub")]
     diff_ids: Vec<String>,
 }
 
@@ -359,7 +447,17 @@ impl Default for RootFs {
 }
 
 #[derive(
-    Builder, Clone, Debug, Default, Deserialize, Eq, Getters, Setters, PartialEq, Serialize,
+    Builder,
+    Clone,
+    Debug,
+    Default,
+    Deserialize,
+    Eq,
+    CopyGetters,
+    Getters,
+    Setters,
+    PartialEq,
+    Serialize,
 )]
 #[builder(
     default,
@@ -399,35 +497,37 @@ mod tests {
     use std::{fs, path::PathBuf};
 
     use super::*;
-    use crate::image::Os;
+    use crate::image::{ANNOTATION_CREATED, ANNOTATION_VERSION};
 
-    fn create_config() -> ImageConfiguration {
-        let configuration = ImageConfigurationBuilder::default()
+    fn create_base_config() -> ConfigBuilder {
+        ConfigBuilder::default()
+            .user("alice".to_owned())
+            .exposed_ports(vec!["8080/tcp".to_owned()])
+            .env(vec![
+                "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin".to_owned(),
+                "FOO=oci_is_a".to_owned(),
+                "BAR=well_written_spec".to_owned(),
+            ])
+            .entrypoint(vec!["/bin/my-app-binary".to_owned()])
+            .cmd(vec![
+                "--foreground".to_owned(),
+                "--config".to_owned(),
+                "/etc/my-app.d/default.cfg".to_owned(),
+            ])
+            .volumes(vec![
+                "/var/job-result-data".to_owned(),
+                "/var/log/my-app-logs".to_owned(),
+            ])
+            .working_dir("/home/alice".to_owned())
+    }
+
+    fn create_base_imgconfig(conf: Config) -> ImageConfigurationBuilder {
+        ImageConfigurationBuilder::default()
             .created("2015-10-31T22:22:56.015925234Z".to_owned())
             .author("Alyssa P. Hacker <alyspdev@example.com>".to_owned())
             .architecture(Arch::Amd64)
             .os(Os::Linux)
-            .config(
-                ConfigBuilder::default()
-                    .user("alice".to_owned())
-                    .exposed_ports(vec!["8080/tcp".to_owned()])
-                    .env(vec![
-                        "PATH=/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin".to_owned(),
-                        "FOO=oci_is_a".to_owned(),
-                        "BAR=well_written_spec".to_owned(),
-                    ])
-                    .entrypoint(vec!["/bin/my-app-binary".to_owned()])
-                    .cmd(vec![
-                        "--foreground".to_owned(),
-                        "--config".to_owned(),
-                        "/etc/my-app.d/default.cfg".to_owned(),
-                    ])
-                    .volumes(vec![
-                        "/var/job-result-data".to_owned(),
-                        "/var/log/my-app-logs".to_owned(),
-                    ])
-                    .working_dir("/home/alice".to_owned())
-                    .build().expect("build config"),
+            .config(conf
             )
             .rootfs(RootFsBuilder::default()
             .diff_ids(vec![
@@ -449,10 +549,27 @@ mod tests {
                 .build()
                 .expect("build history"),
             ])
-            .build()
-            .expect("build configuration");
+    }
 
-        configuration
+    fn create_config() -> ImageConfiguration {
+        create_base_imgconfig(create_base_config().build().expect("config"))
+            .build()
+            .expect("build configuration")
+    }
+
+    /// A config with some additions (labels)
+    fn create_imgconfig_v1() -> ImageConfiguration {
+        let labels = [
+            (ANNOTATION_CREATED, "2023-09-16T19:22:18.014Z"),
+            (ANNOTATION_VERSION, "42.27"),
+        ]
+        .into_iter()
+        .map(|(k, v)| (k.to_owned(), v.to_owned()));
+        let config = create_base_config()
+            .labels(labels.collect::<HashMap<_, _>>())
+            .build()
+            .unwrap();
+        create_base_imgconfig(config).build().unwrap()
     }
 
     fn get_config_path() -> PathBuf {
@@ -472,6 +589,16 @@ mod tests {
         assert_eq!(actual, expected);
     }
 
+    #[test]
+    fn test_helpers() {
+        let config = create_imgconfig_v1();
+        assert_eq!(config.labels_of_config().unwrap().len(), 2);
+        assert_eq!(
+            config.get_config_annotation(ANNOTATION_CREATED).unwrap(),
+            "2023-09-16T19:22:18.014Z"
+        );
+    }
+
     #[test]
     fn load_configuration_from_reader() {
         // arrange
@@ -479,11 +606,11 @@ mod tests {
 
         // act
         let actual = ImageConfiguration::from_reader(&*reader).expect("from reader");
-        println!("{:#?}", actual);
+        println!("{actual:#?}");
 
         // assert
         let expected = create_config();
-        println!("{:#?}", expected);
+        println!("{expected:#?}");
 
         assert_eq!(actual, expected);
     }
@@ -515,9 +642,74 @@ mod tests {
 
         // act
         config.to_writer_pretty(&mut actual).expect("to writer");
+        let actual = String::from_utf8(actual).unwrap();
 
         // assert
-        let expected = fs::read(get_config_path()).expect("read expected");
+        let expected = fs::read_to_string(get_config_path()).expect("read expected");
         assert_eq!(actual, expected);
     }
+
+    #[test]
+    fn save_config_to_string() {
+        // arrange
+        let config = create_config();
+
+        // act
+        let actual = config.to_string_pretty().expect("to string");
+
+        // assert
+        let expected = fs::read_to_string(get_config_path()).expect("read expected");
+        assert_eq!(actual, expected);
+    }
+
+    #[test]
+    fn optional_history_field_absent() {
+        let json = r#"{
+            "architecture": "amd64",
+            "os": "linux",
+            "rootfs": {
+                "type": "layers",
+                "diff_ids": ["sha256:abc123"]
+            }
+        }"#;
+
+        let config: ImageConfiguration =
+            serde_json::from_str(json).expect("deserialize without history");
+        assert!(config.history().is_none());
+    }
+
+    #[test]
+    fn serialize_without_history() {
+        let config = ImageConfigurationBuilder::default()
+            .architecture(Arch::Amd64)
+            .os(Os::Linux)
+            .rootfs(
+                RootFsBuilder::default()
+                    .diff_ids(vec!["sha256:abc123".to_owned()])
+                    .build()
+                    .expect("build rootfs"),
+            )
+            .build()
+            .expect("build config");
+
+        let json = config.to_string().expect("serialize");
+        assert!(!json.contains("history"));
+    }
+
+    #[test]
+    fn builder_without_history() {
+        let config = ImageConfigurationBuilder::default()
+            .architecture(Arch::Amd64)
+            .os(Os::Linux)
+            .rootfs(
+                RootFsBuilder::default()
+                    .diff_ids(vec!["sha256:abc123".to_owned()])
+                    .build()
+                    .expect("build rootfs"),
+            )
+            .build()
+            .expect("build config");
+
+        assert!(config.history().is_none());
+    }
 }
