@@ -1,5 +1,6 @@
 #![deny(missing_docs, warnings)]
 #![doc = include_str!("../README.md")]
+#![allow(clippy::too_long_first_doc_paragraph)]
 
 #[cfg(feature = "distribution")]
 pub mod distribution;
@@ -21,8 +22,8 @@ pub use error::*;
 
 fn from_file<P: AsRef<Path>, T: DeserializeOwned>(path: P) -> Result<T> {
     let path = path.as_ref();
-    let manifest_file = fs::File::open(path)?;
-    let manifest = serde_json::from_reader(&manifest_file)?;
+    let manifest_file = std::io::BufReader::new(fs::File::open(path)?);
+    let manifest = serde_json::from_reader(manifest_file)?;
     Ok(manifest)
 }
 
@@ -38,6 +39,7 @@ fn to_file<P: AsRef<Path>, T: Serialize>(item: &T, path: P, pretty: bool) -> Res
         .create(true)
         .truncate(true)
         .open(path)?;
+    let file = std::io::BufWriter::new(file);
 
     match pretty {
         true => serde_json::to_writer_pretty(file, item)?,
@@ -55,3 +57,10 @@ fn to_writer<W: Write, T: Serialize>(item: &T, writer: &mut W, pretty: bool) ->
 
     Ok(())
 }
+
+fn to_string<T: Serialize>(item: &T, pretty: bool) -> Result<String> {
+    Ok(match pretty {
+        true => serde_json::to_string_pretty(item)?,
+        false => serde_json::to_string(item)?,
+    })
+}
