@@ -1,17 +1,21 @@
 //! [OCI runtime spec](https://github.com/opencontainers/runtime-spec) types and definitions.
+//!
+//! [`Spec`] represents the root object from the specification.
 
 use derive_builder::Builder;
-use getset::Getters;
+use getset::{Getters, MutGetters, Setters};
 use serde::{Deserialize, Serialize};
 use std::{
     collections::HashMap,
     fs,
+    io::{BufReader, BufWriter, Write},
     path::{Path, PathBuf},
 };
 
 use crate::error::{oci_error, OciSpecError, Result};
 
 mod capability;
+mod features;
 mod hooks;
 mod linux;
 mod miscellaneous;
@@ -24,6 +28,7 @@ mod windows;
 
 // re-export for ease of use
 pub use capability::*;
+pub use features::*;
 pub use hooks::*;
 pub use linux::*;
 pub use miscellaneous::*;
@@ -33,15 +38,18 @@ pub use version::*;
 pub use vm::*;
 pub use windows::*;
 
-/// Base configuration for the container.
-#[derive(Builder, Clone, Debug, Deserialize, Getters, PartialEq, Serialize)]
+/// `config.json` file root object.
+#[derive(
+    Builder, Clone, Debug, Deserialize, Getters, MutGetters, Setters, PartialEq, Eq, Serialize,
+)]
+#[serde(rename_all = "camelCase")]
 #[builder(
     default,
     pattern = "owned",
     setter(into, strip_option),
     build_fn(error = "OciSpecError")
 )]
-#[getset(get = "pub")]
+#[getset(get_mut = "pub", get = "pub", set = "pub")]
 pub struct Spec {
     #[serde(default, rename = "ociVersion")]
     ///  MUST be in SemVer v2.0.0 format and specifies the version of the
@@ -89,6 +97,15 @@ pub struct Spec {
     /// the container UTS namespace may be the runtime UTS namespace.
     hostname: Option<String>,
 
+    #[serde(default, skip_serializing_if = "Option::is_none")]
+    /// Specifies the container's domainame as seen by processes running
+    /// inside the container. On Linux, for example, this will
+    /// change the domainame in the container [UTS namespace](http://man7.org/linux/man-pages/man7/namespaces.7.html). Depending on your
+    /// [namespace
+    /// configuration](https://github.com/opencontainers/runtime-spec/blob/master/config-linux.md#namespaces),
+    /// the container UTS namespace may be the runtime UTS namespace.
+    domainname: Option<String>,
+
     #[serde(default, skip_serializing_if = "Option::is_none")]
     /// Hooks allow users to specify programs to run before or after various
     /// lifecycle events. Hooks MUST be called in the listed order.
@@ -130,6 +147,16 @@ pub struct Spec {
     #[serde(default, skip_serializing_if = "Option::is_none")]
     /// VM specifies configuration for Virtual Machine based containers.
     vm: Option<VM>,
+
+    #[serde(default, skip_serializing_if = "Option::is_none")]
+    /// UID mappings used for changing file owners w/o calling chown, fs should support it.
+    /// Every mount point could have its own mapping.
+    uid_mappings: Option<Vec<LinuxIdMapping>>,
+
+    #[serde(default, skip_serializing_if = "Option::is_none")]
+    /// GID mappings used for changing file owners w/o calling chown, fs should support it.
+    /// Every mount point could have its own mapping.
+    gid_mappings: Option<Vec<LinuxIdMapping>>,
 }
 
 // This gives a basic boilerplate for Spec that can be used calling
@@ -143,8 +170,8 @@ impl Default for Spec {
             version: String::from("1.0.2-dev"),
             process: Some(Default::default()),
             root: Some(Default::default()),
-            // Defaults hostname as youki
             hostname: "youki".to_string().into(),
+            domainname: None,
             mounts: get_default_mounts().into(),
             // Defaults to empty metadata
             annotations: Some(Default::default()),
@@ -153,6 +180,8 @@ impl Default for Spec {
             solaris: None,
             windows: None,
             vm: None,
+            uid_mappings: None,
+            gid_mappings: None,
         }
     }
 }
@@ -160,9 +189,8 @@ impl Default for Spec {
 impl Spec {
     /// Load a new `Spec` from the provided JSON file `path`.
     /// # Errors
-    /// This function will return an [OciSpecError::Io](crate::OciSpecError::Io)
-    /// if the spec does not exist or an
-    /// [OciSpecError::SerDe](crate::OciSpecError::SerDe) if it is invalid.
+    /// This function will return an [OciSpecError::Io] if the spec does not exist or an
+    /// [OciSpecError::SerDe] if it is invalid.
     /// # Example
     /// ``` no_run
     /// use oci_spec::runtime::Spec;
@@ -172,16 +200,15 @@ impl Spec {
     pub fn load<P: AsRef<Path>>(path: P) -> Result<Self> {
         let path = path.as_ref();
         let file = fs::File::open(path)?;
-        let s = serde_json::from_reader(&file)?;
+        let reader = BufReader::new(file);
+        let s = serde_json::from_reader(reader)?;
         Ok(s)
     }
 
     /// Save a `Spec` to the provided JSON file `path`.
     /// # Errors
-    /// This function will return an [OciSpecError::Io](crate::OciSpecError::Io)
-    /// if a file cannot be created at the provided path or an
-    /// [OciSpecError::SerDe](crate::OciSpecError::SerDe) if the spec cannot be
-    /// serialized.
+    /// This function will return an [OciSpecError::Io] if a file cannot be created at the provided
+    /// path or an [OciSpecError::SerDe] if the spec cannot be serialized.
     /// # Example
     /// ``` no_run
     /// use oci_spec::runtime::Spec;
@@ -192,7 +219,9 @@ impl Spec {
     pub fn save<P: AsRef<Path>>(&self, path: P) -> Result<()> {
         let path = path.as_ref();
         let file = fs::File::create(path)?;
-        serde_json::to_writer(&file, self)?;
+        let mut writer = BufWriter::new(file);
+        serde_json::to_writer(&mut writer, self)?;
+        writer.flush()?;
         Ok(())
     }
 
@@ -213,6 +242,21 @@ impl Spec {
         Ok(())
     }
 
+    /// Return default rootless spec.
+    /// # Example
+    /// ``` no_run
+    /// use oci_spec::runtime::Spec;
+    ///
+    /// let spec = Spec::rootless(1000, 1000);
+    /// ```
+    pub fn rootless(uid: u32, gid: u32) -> Self {
+        Self {
+            mounts: get_rootless_mounts().into(),
+            linux: Some(Linux::rootless(uid, gid)),
+            ..Default::default()
+        }
+    }
+
     fn canonicalize_path<B, P>(bundle: B, path: P) -> Result<PathBuf>
     where
         B: AsRef<Path>,
@@ -235,7 +279,16 @@ mod tests {
     fn test_canonicalize_rootfs() {
         let rootfs_name = "rootfs";
         let bundle = tempfile::tempdir().expect("failed to create tmp test bundle dir");
-        let rootfs_absolute_path = bundle.path().join(rootfs_name);
+
+        // On macOS, `$TMPDIR` may not point to canonicalized path.
+        // ```
+        // $ echo $TMPDIR; realpath $TMPDIR
+        // /var/folders/_h/j_17023n23s3_50cq_gwhrrc0000gq/T/
+        // /private/var/folders/_h/j_17023n23s3_50cq_gwhrrc0000gq/T
+        // ```
+        let bundle = fs::canonicalize(bundle.path()).expect("failed to canonicalize bundle");
+
+        let rootfs_absolute_path = bundle.join(rootfs_name);
         assert!(
             rootfs_absolute_path.is_absolute(),
             "rootfs path is not absolute path"
@@ -253,7 +306,7 @@ mod tests {
                 .build()
                 .unwrap();
 
-            spec.canonicalize_rootfs(bundle.path())
+            spec.canonicalize_rootfs(&bundle)
                 .expect("failed to canonicalize rootfs");
 
             assert_eq!(
@@ -268,7 +321,7 @@ mod tests {
                 .build()
                 .unwrap();
 
-            spec.canonicalize_rootfs(bundle.path())
+            spec.canonicalize_rootfs(&bundle)
                 .expect("failed to canonicalize rootfs");
 
             assert_eq!(
@@ -284,7 +337,7 @@ mod tests {
             ..Default::default()
         };
         let test_dir = tempfile::tempdir().expect("failed to create tmp test dir");
-        let spec_path = test_dir.into_path().join("config.json");
+        let spec_path = test_dir.keep().join("config.json");
 
         // Test first save the default config, and then load the saved config.
         // The before and after should be the same.
@@ -295,4 +348,114 @@ mod tests {
             "The saved spec is not the same as the loaded spec"
         );
     }
+
+    #[test]
+    fn test_rootless() {
+        const UID: u32 = 1000;
+        const GID: u32 = 1000;
+
+        let spec = Spec::default();
+        let spec_rootless = Spec::rootless(UID, GID);
+        assert!(
+            spec != spec_rootless,
+            "default spec and rootless spec should be different"
+        );
+
+        // Check rootless linux object.
+        let linux = spec_rootless
+            .linux
+            .expect("linux object should not be empty");
+        let uid_mappings = linux
+            .uid_mappings()
+            .clone()
+            .expect("uid mappings should not be empty");
+        let gid_mappings = linux
+            .gid_mappings()
+            .clone()
+            .expect("gid mappings should not be empty");
+        let namespaces = linux
+            .namespaces()
+            .clone()
+            .expect("namespaces should not be empty");
+        assert_eq!(uid_mappings.len(), 1, "uid mappings length should be 1");
+        assert_eq!(
+            uid_mappings[0].host_id(),
+            UID,
+            "uid mapping host id should be as defined"
+        );
+        assert_eq!(gid_mappings.len(), 1, "gid mappings length should be 1");
+        assert_eq!(
+            gid_mappings[0].host_id(),
+            GID,
+            "gid mapping host id should be as defined"
+        );
+        assert!(
+            !namespaces
+                .iter()
+                .any(|ns| ns.typ() == LinuxNamespaceType::Network),
+            "rootless spec should not contain network namespace type"
+        );
+        assert!(
+            namespaces
+                .iter()
+                .any(|ns| ns.typ() == LinuxNamespaceType::User),
+            "rootless spec should contain user namespace type"
+        );
+        assert!(
+            linux.resources().is_none(),
+            "resources in rootless spec should be empty"
+        );
+
+        // Check rootless mounts.
+        let mounts = spec_rootless.mounts.expect("mounts should not be empty");
+        assert!(
+            !mounts.iter().any(|m| {
+                if m.destination().to_string_lossy() == "/dev/pts" {
+                    return m
+                        .options()
+                        .clone()
+                        .expect("options should not be empty")
+                        .iter()
+                        .any(|o| o == "gid=5");
+                } else {
+                    false
+                }
+            }),
+            "gid=5 in rootless should not be present"
+        );
+        let sys_mount = mounts
+            .iter()
+            .find(|m| m.destination().to_string_lossy() == "/sys")
+            .expect("sys mount should be present");
+        assert_eq!(
+            sys_mount.typ(),
+            &Some("none".to_string()),
+            "type should be changed in sys mount"
+        );
+        assert_eq!(
+            sys_mount
+                .source()
+                .clone()
+                .expect("source should not be empty in sys mount")
+                .to_string_lossy(),
+            "/sys",
+            "source should be changed in sys mount"
+        );
+        assert!(
+            sys_mount
+                .options()
+                .clone()
+                .expect("options should not be empty in sys mount")
+                .iter()
+                .any(|o| o == "rbind"),
+            "rbind option should be present in sys mount"
+        );
+
+        // Check that some other objects have same values.
+        assert!(spec.process == spec_rootless.process);
+        assert!(spec.root == spec_rootless.root);
+        assert!(spec.hooks == spec_rootless.hooks);
+        assert!(spec.uid_mappings == spec_rootless.uid_mappings);
+        assert!(spec.gid_mappings == spec_rootless.gid_mappings);
+    }
 }
